#include "cmac.h"

int AES_ECB(unsigned char *out, unsigned char *in, unsigned int len, unsigned char *key, int keyLen){
    if (keyLen == 16) {
        aes_128_encrypt(in, out, key);
    }else if(keyLen == 24){
        aes_192_encrypt(in, out, key);
    }else if (keyLen == 32){
        aes_256_encrypt(in, out, key);
    }else{
        aes_128_encrypt(in, out, key);
    }
    return 0;
}


/*
@param[in]      data     The buffer to 64bits data to left shift one;
@param[out]    data     save result to data

@param[out]    data     save result to data

Sequence:
bit63              ...     bit0

Byte0             ...     Byte7
bit7... bit0               bit7 ... bit0
*/
void R64_ShiftOneBitLeft(unsigned char *data)
{
    int first4bytesMSBset = 0;
    unsigned int p;

    if(data[4] & 0x80)
    {
        first4bytesMSBset = 1;
    }

    p = (data[0] << 24) | (data[1] << 16) | (data[2] << 8) | data[3];
    p <<= 1;

    p = p | first4bytesMSBset;

    data[0] = (p >> 24) & 0xFF;
    data[1] = (p >> 16) & 0xFF;
    data[2] = (p >> 8) & 0xFF;
    data[3] = p & 0xFF;

    p = (data[4] << 24) | (data[5] << 16) | (data[6] << 8) | data[7];
    p <<= 1;

    data[4] = (p >> 24) & 0xFF;
    data[5] = (p >> 16) & 0xFF;
    data[6] = (p >> 8) & 0xFF;
    data[7] = p & 0xFF;
}


/*
@param[in]      data     The buffer to << 1;
@param[out]    data     save result to data

Sequence:
bit127              ...     bit0

Byte0             ...     Byte15
bit7... bit0               bit7 ... bit0
*/
void R128_ShiftOneBitLeft(unsigned char *data)
{
    int first8bytesMSBset = 0;

    if(data[8] & 0x80)
    {
        first8bytesMSBset = 1;
    }

    R64_ShiftOneBitLeft(data);
    data[7] |= first8bytesMSBset;

    R64_ShiftOneBitLeft(data + 8);
}

/*
@param[in]   a     a[i] = a[i] ^ b[i]
@param[in]   b
@param[in]   len   The length for XOR
*/
void XOR_Arrays(unsigned char *a, const unsigned char *b, unsigned int len)
{
    unsigned int i;

    for(i = 0; i < len; i++)
    {
        a[i] = a[i] ^ b[i];
    }
}



/*
   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   +                    Algorithm Generate_Subkey                      +
   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   +                                                                   +
   +   Input    : K (128-bit key)                                      +
   +   Output   : K1 (128-bit first subkey)                            +
   +              K2 (128-bit second subkey)                           +
   +-------------------------------------------------------------------+
   +                                                                   +
   +   Constants: const_Zero is 0x00000000000000000000000000000000     +
   +              const_Rb   is 0x00000000000000000000000000000087     +
   +   Variables: L          for output of AES-128 applied to 0^128    +
   +                                                                   +
   +   Step 1.  L := AES-128(K, const_Zero);                           +
   +   Step 2.  if MSB(L) is equal to 0                                +
   +            then    K1 := L << 1;                                  +
   +            else    K1 := (L << 1) XOR const_Rb;                   +
   +   Step 3.  if MSB(K1) is equal to 0                               +
   +            then    K2 := K1 << 1;                                 +
   +            else    K2 := (K1 << 1) XOR const_Rb;                  +
   +   Step 4.  return K1, K2;                                         +
   +                                                                   +
   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/
/*
@param[out]  S    Output S
@param[out]  K1   Output K1
@param[out]  K2   Output K2
@param[in]    K     Input K
@return       0               No error
@return       !0             error
*/
int CMAC_SubKey_AES(unsigned char *S, unsigned char *K1, unsigned char *K2, const unsigned char *K, int keyLen)
{
    unsigned char data[16] = {0};

    /*For AES algorithm,b=128*/
    const unsigned char Rb128[16] =
    {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x87
    };

    AES_ECB(data, data, 16, (unsigned char *)K, keyLen);

    if(S != NULL)
        memcpy(S, data, 16); /*S*/

    if(data[0] & 0x80) /*MSB=1*/
    {
        R128_ShiftOneBitLeft(data);

        XOR_Arrays(data, Rb128, 16);
    }
    else
    {
        R128_ShiftOneBitLeft(data);
    }

    if(K1 != NULL)
        memcpy(K1, data, 16); /*K1*/

    if(data[0] & 0x80) /*MSB=1*/
    {
        R128_ShiftOneBitLeft(data);

        XOR_Arrays(data, Rb128, 16);
    }
    else
    {
        R128_ShiftOneBitLeft(data);
    }

    if(K2 != NULL)
        memcpy(K2, data, 16); /*K2*/

    memset(data, 0, sizeof(data));

    return 0;
}


/*
@param[out] output              The buffer sotre MAC Output(16bytes)
@param[in] message            The buffer of Message.
@param[in] message_bytes   The length of Message,
@param[in] key                   An AES key

@return       0               No error
@return       !0             error
*/
int CMAC_AES(unsigned char *cmac, const unsigned char *m, unsigned int mLen, const unsigned char *key, int keyLen)
{
    int i;

    unsigned char blockData[16] = {0};

    unsigned char K1[16] = {0};
    unsigned char K2[16] = {0};

    int alignLen = 0;
    int alignPadFlag = 0;

    unsigned char *mBuffer = NULL;

    if((mLen % 16 == 0) && (mLen))
    {
        alignLen = mLen;
        mBuffer = (unsigned char *)m;
    }
    else/*padding*/
    {
        if(mLen)
        {
            alignLen =  ((mLen + 15) >> 4) << 4;
            mBuffer = calloc(1, alignLen);
            memcpy(mBuffer, m, mLen);
        }
        else
        {
            alignLen = 16;
            mBuffer = calloc(1, alignLen);
        }

        mBuffer[mLen] = 0x80;
        alignPadFlag = 1;  /*padding flag*/
    }

    CMAC_SubKey_AES(NULL, K1, K2, key, keyLen);

    for(i = 0; i < alignLen; i += 16)
    {
        XOR_Arrays(blockData, mBuffer + i, 16);

        if((i + 16) == alignLen) /*last block*/
        {
            if(alignPadFlag)
            {
                XOR_Arrays(blockData, K2, 16);
            }
            else
            {
                XOR_Arrays(blockData, K1, 16);
            }
        }

        AES_ECB(blockData, blockData, 16, (unsigned char *)key, keyLen);
    }

    memcpy(cmac, blockData, 16);

    if(alignPadFlag)
    {
        memset(mBuffer, 0, alignLen);

        free(mBuffer);
    }

    return 0;
}
